# File cnnfunctions.py:
# Defines layer operations of
# convolution, activation, pooling, fully-connected, and softmax,
# their backpropagation counterparts,
# and implements the ADADELTA gradient descent method.

import numpy as np

def conv(inT, weight, outT, forward=True, bias=None, inL=None, WDeriv=None, bDeriv=None):
    inShape = inT.shape
    inHeight, inWidth = inShape[:2]

    WShape = weight.shape
    WHeight, WWidth = WShape[:2]

    YBorder = (WHeight-1)/2
    XBorder = (WWidth-1)/2

    if forward:
      outT.fill(0)
    else:
      inT.fill(0)

    for y in np.arange(WHeight):
        yOffset = y-YBorder
        inTop = max(-yOffset, 0)
        outTop = max(yOffset, 0)
        inBottom = inHeight+min(-yOffset, 0)
        outBottom = inHeight+min(yOffset, 0)

        for x in np.arange(WWidth):
            xOffset = x-XBorder
            inLeft = max(-xOffset, 0)
            outLeft = max(xOffset, 0)
            inRight = inWidth+min(-xOffset, 0)
            outRight = inHeight+min(xOffset, 0)

            weightPixel = weight[y, x]
            inSec = inT[inTop:inBottom, inLeft:inRight]
            outSec = outT[outTop:outBottom, outLeft:outRight]

            if forward:
              outSec += np.einsum('ijkm,kl', inSec, weightPixel)
            else:
              inSec += np.einsum('ijlm,kl', outSec, weightPixel)
              np.einsum('ijkm,ijlm', inL[inTop:inBottom, inLeft:inRight], outSec, out=WDeriv[y, x])
    
    if forward:
        outT += bias
    else:
        np.sum(outT, axis=(0, 1, 3), out=bDeriv[:, 0])

def act(inL, f, outL):
    outL[:] = f(inL)
    
def actPrime(dJdinL, inL, fPrime, outL, dJdoutL):
    dJdinL[:] = dJdoutL * fPrime(inL, outL)

def pool(inL, factor, outL):
    inShape = inL.shape
    outL[:] = np.average(
      inL.reshape(
        (inShape[0]/factor, factor, inShape[1]/factor, factor, inShape[2], inShape[3])), 
      axis=(1, 3))

def scale(dJdinL, factor, dJdoutL):
    dJdinL[:] = np.kron(dJdoutL, np.full((factor, factor, 1, 1), factor**-2))

def fc(inL, weight, bias, outL):
    inDims = np.arange(inL.ndim-1)
    outL[:] = np.tensordot(weight, inL, axes=(inDims, inDims))+bias

def fcTranspose(dJdinL, inL, weight, dJdoutL, dJdWeight, dJdbias):
    outDims = dJdoutL.ndim-1
    outAxes = np.arange(outDims)
    WOutAxes = outAxes-outDims
    dJdinL[:] = np.tensordot(weight, dJdoutL, axes=(WOutAxes, outAxes))
    dJdWeight[:] = np.tensordot(inL, dJdoutL, axes=(-1, -1))
    dJdbias[:] = np.expand_dims(np.sum(dJdoutL, axis=-1), -1)

def softmax(inL, outL):
    np.exp(inL-np.amax(inL, axis=0), out=outL)
    outL /= np.sum(outL, axis=0)

def softmaxPrime(dJdinL, outL, dJdoutL):
    a = dJdoutL*outL
    dJdinL[:] = a-(outL*np.sum(a, axis=0))

def findRMS(param, paramMS, m, e):
    paramMS[:] = m*paramMS + (1-m)*param*param
    return np.sqrt(paramMS+e)

def adadelta(dJda, dJdaMS, aStep, aStepMS, m, e):
    dJdaRMS = findRMS(dJda, dJdaMS, m, e)
    aStepRMS = findRMS(aStep, aStepMS, m, e)
    aStep[:] = -aStepRMS*dJda/dJdaRMS
