# File mycnn.py:
# Defines the convolutional network class
# Implements forward and backpropagation at the layer level

import numpy as np
import cnnfunctions # See cnnfunctions.py

class CNN:
    
    def __init__(self, inSize, inNeurons, batchSize, lossFunc, lossPrime):

        inShape = (inSize, inSize, inNeurons, batchSize)

        self.L = [np.ndarray(inShape)]
        self.W = []
        self.b = []

        self.dJdL = [np.ndarray(inShape)]
        self.dJdW = []
        self.dJdb = []

        self.numLayers = 0

        self.dJdWMS = []
        self.WStep = []
        self.WStepMS = []

        self.dJdbMS = []
        self.bStep = []
        self.bStepMS = []

        self.momentum = 0.9
        self.epsilon = 1e-4
        self.decay = 1e-4

        self.params = []

        self.batchSize = batchSize
        self.lossFunc = lossFunc
        self.lossPrime = lossPrime

    def randomize(self):
        for l in range(self.numLayers):
            param = self.params[l]['type']
            normFactor = 1
            if param == 'conv':
                normFactor = kernel*kernel*lastNeurons
            elif param == 'fc':
                normFactor = np.prod(lastShape)/self.batchSize
        
    def addLayer(self, param):

        LType = param['type']
        lastShape = self.L[-1].shape
        
        # Default values
        LShape = list(lastShape)
        WShape, bShape = 0, 0

        normFactor = 1
        
        if LType == 'conv':
            kernel = param['kernel']
            neurons = param['neurons']
            lastNeurons = lastShape[2]
            normFactor = kernel*kernel*lastNeurons
            assert kernel%2 == 1
            LShape[2] = neurons
            WShape = (kernel, kernel, lastNeurons, neurons)
            bShape = (neurons, 1)
            
        elif LType == 'act':
            pass

        elif LType == 'pool':
            factor = param['factor']
            assert lastShape[0]%factor == 0
            LShape[0] /= factor
            assert lastShape[1]%factor == 0
            LShape[1] /= factor

        elif LType == 'fc':
            shape = param['shape']
            LShape = shape+(self.batchSize,)
            WShape = lastShape[:-1]+shape
            bShape = shape+(1,)
            normFactor = np.prod(lastShape)/self.batchSize

        elif LType == 'softmax':
            assert len(lastShape)==2

        else:
            print "Invalid layer."

        self.L.append(np.ndarray(LShape))
        self.W.append(np.random.random(WShape)*2/normFactor)
        self.b.append(np.random.random(bShape)-0.5)

        self.dJdL.append(np.ndarray(LShape))
        self.dJdW.append(np.ndarray(WShape))
        self.dJdb.append(np.ndarray(bShape))

        # MS = mean square
        self.dJdWMS.append(np.zeros(WShape))
        self.WStep.append(np.zeros(WShape))
        self.WStepMS.append(np.zeros(WShape))

        self.dJdbMS.append(np.zeros(bShape))
        self.bStep.append(np.zeros(bShape))
        self.bStepMS.append(np.zeros(bShape))
        
        self.params.append(param)

        self.numLayers += 1
        
    def forward(self, X):
        self.L[0][:] = X
        for l in range(self.numLayers):
            param = self.params[l]
            LType = param['type']
            if LType == 'conv':
                cnnfunctions.conv(self.L[l], self.W[l], self.L[l+1], bias=self.b[l])
            elif LType == 'pool':
                cnnfunctions.pool(self.L[l], param['factor'], self.L[l+1])
            elif LType == 'act':
                cnnfunctions.act(self.L[l], param['f'], self.L[l+1])
            elif LType == 'fc': 
                cnnfunctions.fc(self.L[l], self.W[l], self.b[l], self.L[l+1])
            elif LType == 'softmax':
                cnnfunctions.softmax(self.L[l], self.L[l+1])
            else:
                raise ValueError('Invalid layer type.')
            
    def backward(self, dJdyHat):
        self.dJdL[-1][:] = dJdyHat
        for m in range(self.numLayers, 0, -1):
            l = m-1
            param = self.params[l]
            LType = param['type']
            if LType == 'conv':
                cnnfunctions.conv(self.dJdL[l], self.W[l], self.dJdL[l+1], forward=False, inL=self.L[l], WDeriv=self.dJdW[l], bDeriv=self.dJdb[l])
            elif LType == 'pool':
                cnnfunctions.scale(self.dJdL[l], param['factor'], self.dJdL[l+1]) 
            elif LType == 'act':
                cnnfunctions.actPrime(self.dJdL[l], self.L[l], param['fPrime'], self.L[l+1], self.dJdL[l+1])
            elif LType == 'fc':
                cnnfunctions.fcTranspose(self.dJdL[l], self.L[l], self.W[l], self.dJdL[l+1], self.dJdW[l], self.dJdb[l])
            elif LType == 'softmax':
                cnnfunctions.softmaxPrime(self.dJdL[l], self.L[l+1], self.dJdL[l+1])
            else:
                raise ValueError('Invalid layer type.')
                break
            
            cnnfunctions.adadelta(self.dJdW[l], self.dJdWMS[l], self.WStep[l], self.WStepMS[l], self.momentum, self.epsilon)
            self.W[l] += self.WStep[l]
            self.W[l] *= 1-self.decay

            cnnfunctions.adadelta(self.dJdb[l], self.dJdbMS[l], self.bStep[l], self.bStepMS[l], self.momentum, self.epsilon)
            self.b[l] += self.bStep[l]
            self.b[l] *= 1-self.decay

    def train(self, getBatches):
        iterations = 0
        for X, y, t in getBatches(self.batchSize):
            self.forward(X)
            yHat = self.L[-1]
            if t:
                dJdyHat = self.lossPrime(y, yHat)
                self.backward(dJdyHat)
            else:
                pred = np.argmax(yHat, axis=0)
                actual = np.argmax(y, axis=0)
                numRight = np.sum(pred==actual)
                yield numRight

